# -*- coding: utf-8 -*-

"""Top-level package for SEAMM Widgets."""

__author__ = """Paul Saxe"""
__email__ = 'psaxe@vt.edu'
__version__ = '0.2.1'

# Bring up the classes so that they appear to be directly in
# the seamm_widgets package.

from seamm_widgets.mousewheel_support import MousewheelSupport  # noqa: F401
from seamm_widgets.scrolled_frame import ScrolledFrame  # noqa: F401
from seamm_widgets.scrolled_columns import ScrolledColumns  # noqa: F401
from seamm_widgets.labeled_widget import LabeledWidget, align_labels  # noqa: F401, E501
from seamm_widgets.labeled_entry import LabeledEntry  # noqa: F401
from seamm_widgets.labeled_combobox import LabeledCombobox  # noqa: F401
from seamm_widgets.unit_entry import UnitEntry  # noqa: F401
from seamm_widgets.unit_combobox import UnitCombobox  # noqa: F401

default_units = {
    '[length]': ['Å', 'pm', 'nm', 'um'],
    '[length] ** 3': ['Å**3', 'pm**3', 'nm**3', 'um**3'],
    '[mass]': ['g', 'kg', 'tonne', 'lb', 'ton'],
    '[substance]': ['mol'],
    '[temperature]': ['K', 'degC', 'degF', 'degR'],
    '[mass] / [length] ** 3': ['g/mL', 'kg/L', 'kg/m**3',
                               'g/mol/Å**3'],  # nopep8
    '[time]': ['fs', 'ps', 'ns', 'us', 'ms', 's'],
    '[length] ** 2 * [mass] / [substance] / [time] ** 2':
        ['kcal/mol', 'kJ/mol'],  # nopep8
    '[length] * [mass] / [substance] / [time] ** 2':
        ['kcal/mol/Å', 'kJ/mol/Å', 'eV/Å'],  # nopep8
    '[length] * [mass] / [time] ** 2': ['kcal/mol/Å', 'kJ/mol/Å',
                                        'eV/Å'],  # nopep8
    '[length] / [time] ** 2': ['m/s**2', 'ft/s**2', 'Å/fs**2'],
    '[mass] / [length] / [time] ** 2': ['Pa', 'atm', 'bar', 'psi', 'ksi'],
}



# -*- coding: utf-8 -*-

"""A ttk-based widget for columns of widgets, with fixed titles and scrolling

This widgets has two areas: a row of titles across the top and a scrolled frame
below it. It is used to make a table of widgets with fixed column headers.
"""

import seamm_widgets as sw
import tkinter as tk
from tkinter import ttk


class ScrolledColumns(ttk.Frame):

    def __init__(self, parent, *args, **kwargs):
        """Initialize the widget
        """

        class_ = kwargs.pop('class_', 'MLabeledEntry')
        super().__init__(parent, class_=class_)

        columns = kwargs.pop('columns', [])
        self.min_sizes = kwargs.pop('minsize', [])
        self._after_id = None

        # list (vector) of the header widgets, _ncolumns long
        self._header_widgets = []

        # list of lists (row x column) of widgets in the table
        self._widgets = []

        # Create the two subframes, linking them both to the
        # horizontal scrollbar at the bottom
        # self.headers = ttk.Frame(self)
        self.table = sw.ScrolledFrame(
            self, scroll_vertically=True, borderwidth=2, relief=tk.SUNKEN
        )
        self.headers = sw.ScrolledFrame(
            self,
            scroll_vertically=False,
            borderwidth=2,
            relief=tk.RAISED,
            xscrollbar=self.table.xscrollbar
        )

        # and patch up the horizontal scrolling
        self.table.xscrollbar['command'] = self.xview

        self.headers.grid(row=0, column=0, sticky=tk.EW)
        self.table.grid(row=1, column=0, sticky=tk.NSEW)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)

        # Put in the column titles if given
        col = 0
        header = self.headers.interior()
        for item in columns:
            if isinstance(item, str):
                item = ttk.Label(header, text=item)
            item.grid(row=0, column=col)
            col += 1
            self._header_widgets.append(item)

    def cell(self, row, column, value=None):
        """Return or set the widget at the given cell
        """
        if value is None:
            try:
                result = self._widgets[row][column]
            except Exception:
                raise
            return result
        else:
            if column >= self.ncolumns:
                # pad the header and each row with None's
                extra = [None] * (column - self.ncolumns + 1)
                self._header_widgets.extend(extra)
                for row_of_widgets in self._widgets:
                    row_of_widgets.extend(extra)
            if row >= self.nrows:
                # add rows 'till we get there
                extra = [None] * self.ncolumns
                for i in range(self.nrows, row + 1):
                    self._widgets.append(extra)

            if isinstance(value, str):
                value = ttk.Label(self.table.interior(), text=value)

            value.grid(row=row, column=column)
            self._widgets[row][column] = value

            self._update_widths()

    def delete_row(self, index):
        for item in self._widgets[index]:
            if item is not None:
                item.destroy()
        del self._widgets[index]

        self._update_widths()

    def delete_column(self, index):
        for row in self._widgets:
            item = row[index]
            if item is not None:
                item.destroy()
            del row[index]

        self._update_widths()

    # Provide matrix-like access to the widgets to make
    # the code cleaner

    def __getitem__(self, index):
        """Allow [row,column] access to the widgets!"""
        if isinstance(index, tuple):
            row, column = index
            return self.cell(row, column)
        else:
            raise Exception("Row and column indices are required")

    def __setitem__(self, index, value):
        """Allow x[row,column] access to the data"""
        if isinstance(index, tuple):
            row, column = index
            self.cell(row, column, value)
        else:
            raise Exception("Row and column indices are required")

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if isinstance(key, tuple):
            row, column = key
            widget = self.cell(row, column)
            widget.destroy()
            self.cell(row, column, None)

            self._update_widths()
        else:
            raise Exception("Row and column indices are required")

    @property
    def nrows(self):
        return len(self._widgets)

    @property
    def ncolumns(self):
        return len(self._header_widgets)

    def _update_widths_now(self):
        """Force the update of the column widths to happen now"""
        self._update_widths(when='now')

    def _update_widths(self, when='later'):
        """Make the column widths of header and table identical"""

        if when == 'later':
            if self._after_id is None:
                self._after_id = self.after_idle(self._update_widths_now)
            return

        self.after_cancel(self._after_id)
        self._after_id = None

        header = self.headers.interior()
        table = self.table.interior()

        # Remove any widths currently set
        for column in range(0, self.ncolumns):
            header.columnconfigure(column, minsize=0)
            table.columnconfigure(column, minsize=0)

        # Let everything settle and finds its size
        header.update_idletasks()

        # And make the sizes equal
        for column in range(0, self.ncolumns):
            # bbox returns (x, y, w, h) so we want the third item
            w1 = header.grid_bbox(column, 0)[2]
            w2 = table.grid_bbox(column, 0)[2]
            if w1 < w2:
                header.columnconfigure(column, minsize=w2)
            else:
                table.columnconfigure(column, minsize=w1)

        # ensure that the header area is large enough
        h = header.grid_bbox(0, 0)[3]
        self.headers.height = h + 5

    def interior(self):
        """Where the user packs widgets"""
        return self.table.interior()

    def xview(self, *args):
        """Connect the subwidgets to the single scrollbar"""
        self.headers.canvas.xview(*args)
        self.table.canvas.xview(*args)



# -*- coding: utf-8 -*-

"""Labeled combobox widget with units.

The goal of these widgets is twofold: to make it easier for developers
to implement dialogs with compound widgets, and to naturally
standardize the user interface presented to the user.
"""

import logging
from seamm_util import ureg, Q_, units_class  # noqa: F401
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)

options = {
    'unitcombobox': {
        'as_quantity': 'as_quantity',
    },
    'units':
        {
            'class_': 'class_',
            'cursor': 'cursor',
            'exportselection': 'exportselection',
            'unitsheight': 'height',
            'unitsjustify': 'justify',
            'postcommand': 'postcommand',
            'style': 'style',
            'unitstakefocus': 'takefocus',
            'variable': 'textvariable',
            'unitsvalidate': 'validate',
            'unitsvalidatecommand': 'validatecommand',
            'unitswidth': 'width',
            'unitsxscrollcommand': 'xscrollcommand',
        }
}


class UnitCombobox(sw.LabeledCombobox):

    def __init__(self, parent, *args, **kwargs):
        """Initialize the instance
        """
        class_ = kwargs.pop('class_', 'MUnitCombobox')
        super().__init__(parent, class_=class_)

        interior = self.interior

        # unitcombobox options
        self.as_quantity = kwargs.pop('as_quantity', False)

        # units combobox
        unitsheight = kwargs.pop('unitsheight', 7)
        unitswidth = kwargs.pop('unitswidth', 10)
        unitsstate = kwargs.pop('unitsstate', 'readonly')

        self.units = ttk.Combobox(
            interior, height=unitsheight, width=unitswidth, state=unitsstate
        )
        self.units.grid(row=0, column=0, sticky=tk.EW)

        # interior frame
        self.interior = ttk.Frame(interior)
        self.interior.grid(row=0, column=1, sticky=tk.NSEW)

        self.config(**kwargs)

    @property
    def value(self):
        return self.get()

    @value.setter
    def value(self, value):
        self.set(value)

    def show(self, *args):
        """Show only the specified subwidgets.
        'all' or no arguments reverts to showing all"""

        super().show(*args)

        show_all = (len(args) == 0 or args[0] == 'all')

        if show_all or 'units' in args:
            self.units.grid(row=0, column=2, sticky=tk.W)
        else:
            self.units.grid_forget()

    def set(self, value, unit_string=None):
        """Set the the value and units"""

        if value is None:
            return

        # the value may have units or be a plain value
        if isinstance(value, units_class):
            self.combobox.set(value.magnitude)

            dimensionality = value.dimensionality
            current_units = self.units.cget('values')
            if len(current_units) > 0:
                for unit in current_units:
                    if Q_(unit).dimensionality != dimensionality:
                        self.units.configure(values=[])
                        current_units = []
                        break

            if len(current_units) == 0:
                self.set_units([*sw.default_units[str(dimensionality)], ''])
                self.units.set('{0.units:~}'.format(value).replace(' ', ''))
        elif unit_string:
            self.combobox.set(value)

            dimensionality = Q_(unit_string).dimensionality
            current_units = self.units.cget('values')
            if len(current_units) > 0:
                for unit in current_units:
                    if Q_(unit).dimensionality != dimensionality:
                        self.units.configure(values=[])
                        current_units = []
                        break

            if len(current_units) == 0:
                self.set_units([*sw.default_units[str(dimensionality)], ''])
                self.units.set(unit_string)
        else:
            self.combobox.set(value)
            self.set_units('all')
            self.units.set('')

    def get(self):
        """return the current value with units"""
        value = self.combobox.get()
        if value in self.combobox.cget('values'):
            return value
        else:
            unit = self.units.get()
            if unit == '':
                return value
            elif self.as_quantity:
                try:
                    magnitude = float(value)
                    return Q_(magnitude, unit)
                except:  # noqa: E722
                    return (value, unit)
            else:
                return (value, unit)

    def set_units(self, values=None):
        if values is None:
            dimensionality = str(self.get().dimensionality)
            self.units.config(values=sw.default_units[dimensionality])
        elif values == 'all':
            tmp = ['']
            for key in sw.default_units:
                tmp += sw.default_units[key]
            self.units.config(values=tmp)
        else:
            self.units.config(values=values)

    def config(self, **kwargs):
        """Set the configuration of the megawidget"""
        unitcombobox = options['unitcombobox']
        units = options['units']

        # cannot modify kwargs while iterating over it...
        keys = [*kwargs.keys()]
        for k in keys:
            if k in unitcombobox and unitcombobox[k] in self.__dict__:
                v = kwargs.pop(k)
                self.__dict__[unitcombobox[k]] = v
            elif k in units:
                v = kwargs.pop(k)
                self.units.config(**{units[k]: v})

        # having removed our options, pass rest to parent
        super().config(**kwargs)



# -*- coding: utf-8 -*-

"""Labeled combobox widget.

The goal of these widgets is twofold: to make it easier for developers
to implement dialogs with compound widgets, and to naturally
standardize the user interface presented to the user.
"""

import logging
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)

options = {
    'combobox':
        {
            'class_': 'class_',
            'cursor': 'cursor',
            'exportselection': 'exportselection',
            'justify': 'justify',
            'height': 'height',
            'postcommand': 'postcommand',
            'state': 'state',
            'style': 'style',
            'takefocus': 'takefocus',
            'variable': 'textvariable',
            'values': 'values',
            'width': 'width'
        },
}


class LabeledCombobox(sw.LabeledWidget):

    def __init__(self, parent, *args, **kwargs):
        """Initialize the instance
        """
        class_ = kwargs.pop('class_', 'MLabeledCombobox')
        super().__init__(parent, class_=class_)

        interior = self.interior

        # combobox
        height = kwargs.pop('height', 7)
        width = kwargs.pop('width', 20)
        state = kwargs.pop('state', 'normal')

        self.combobox = ttk.Combobox(
            interior, height=height, width=width, state=state
        )
        self.combobox.grid(row=0, column=0, sticky=tk.EW)

        # interior frame
        self.interior = ttk.Frame(interior)
        self.interior.grid(row=0, column=1, sticky=tk.NSEW)

        interior.columnconfigure(0, weight=1)

        self.config(**kwargs)

    @property
    def value(self):
        return self.get()

    @value.setter
    def value(self, value):
        self.set(value)

    def show(self, *args):
        """Show only the specified subwidgets.
        'all' or no arguments reverts to showing all"""

        super().show(*args)

        show_all = (len(args) == 0 or args[0] == 'all')

        if show_all or 'combobox' in args:
            self.combobox.grid(row=0, column=1, sticky=tk.EW)
        else:
            self.combobox.grid_forget()

    def set(self, value):
        """Set the value of the widget"""

        if value is None:
            return

        self.combobox.set(value)

    def get(self):
        """return the current value"""
        value = self.combobox.get()
        return value

    def config(self, **kwargs):
        """Set the configuration of the megawidget"""

        # our options that we deal with
        combobox = options['combobox']

        # cannot modify kwargs while iterating over it...
        keys = [*kwargs.keys()]
        for k in keys:
            if k in combobox:
                v = kwargs.pop(k)
                self.combobox.config(**{combobox[k]: v})

        # having removed our options, pass rest to parent
        super().config(**kwargs)
